/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2007 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */
/*
 *  This file contains an ISA-portable PIN tool for tracing memory accesses.
 */

#include <stdio.h>
#include "pin.H"
#include "instlib.H"
#include <pthread.h>
#include <stdlib.h>
#include <setjmp.h>
#include "format.h"

//#define STDOUT
#define PIPE 1
#define MAX_REFS 20000000 // we want to limit the size of the log files so we will limit trace files to 2M lines
#define MAX_THREADS 2048
#define TRACE_DIR "./TRACE_DIR"
#define SKIP_DEPTH    0 //4 /* not used */
#define TRACE_DEPTH 10 //1000 /* maximum length of stack sig */
#define PRSD_PIPE "/home/sbudanu/prsd_pipe"  //"tmp.trace" //"/home/sbudanu/prsd_pipe"
#define SP_BP_DIFF 100000
#define BIN_MARGIN (unsigned long)(0x0000000000FFFFFFUL) //16MB
#define ABS(a) (a>=0?a:-a)
#define RANGE(x) (unsigned long)((unsigned long)x & (unsigned long)(0xFFFFFFFFFF000000UL))
#define TRACE_TAG 0xBEADBEAD

#if 0
#define LOAD_OUTPUT_FORMAT "ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u l addr:%12p  %u\n", \
                       ip,     ebp,     sp,     (void*)sign,      thread_id, addr,     sincelast
#define STORE_OUTPUT_FORMAT "ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u s addr:%12p  %u\n", \
                       ip,     ebp,     sp,     (void*)sign,      thread_id, addr,     sincelast
#endif

using namespace std;
FILE* fpipe = NULL;
FILE* logfptr = NULL;
long prev_sign[MAX_THREADS];
//FILE * trace;

//application's main routine's address is stored 
AFUNPTR mainFunAddr = 0;

INSTLIB::ICOUNT icount;
UINT32 icount_lastmem;
UINT32 sincelast;

jmp_buf jmp_buffer;
UINT32 memrefs; // number of memory accesses so far

#ifdef MULTI_FILE //Sandeep: need not save to different files. Just print to stdout and pipe to prsd_comp binary
FILE * fptr[MAX_THREADS];
char fileName[256];
#endif

enum
{
 INS_TYPE_LOAD = 0x11223344AABBCCDD,
 INS_TYPE_STORE = 0x556677889900EEFF
};

int main(int argc, char *argv[]);

#include "myqueue.cpp"
LIST<long> signList;

INSTLIB::CONTROL control;
bool doinst;
PIN_LOCK lock;

#if defined(__x86_64__)
__inline__ static void **NextStackFrame(void **old_sp) 
{
  void **new_sp = (void **) *old_sp;

  //fprintf(stdout, "##new-sp:%p old_sp:%p##\n",(void*)new_sp, (void*)old_sp);

  if (new_sp <= old_sp) 
  {
   //fprintf(stdout, "11111\n");
   return NULL;
  }
  if ((uintptr_t)new_sp & (sizeof(void *) - 1)) 
  {
   //fprintf(stdout, "22222\n");
   return NULL;
  }
  if ((uintptr_t)new_sp - (uintptr_t)old_sp > 10000) 
  {
   //fprintf(stdout, "33333\n");
   return NULL;
  }
  return new_sp;
}

//int* do_stack_unwind (void** sfp)
/*
 * perform stackwalk and compute the signature
 */
int getSignature(void** pc, void** sfp, long* sign)
{
  int n = 0;
  void** new_sfp = NULL;

  if(!sign || !pc || !sfp)  
   return -1;

  *sign = (long)pc;
  signList.addTail(*sign);	

  while (sfp && n < TRACE_DEPTH-1) 
   {
    if (*(sfp+1) == (void *)0) //return address is NULL
    {
      break;
    }
    
    //fprintf(stdout, "*%16p* ",(long*)*(sfp+1)); 
    (*sign) ^= (long)(*(sfp+1)); //XOR PC and return addresses to compute unique signature
    //fprintf(stdout, "#Sign: %16p# ",(long*)(*sign)); 
    signList.addTail((long)(*(sfp+1)));	
    new_sfp = NextStackFrame(sfp);
    
    if (!new_sfp) 
    {
      break;
    }
    
      n++;
      sfp = new_sfp;
   }

  //fprintf(stdout, "\n");

  return 0;
}
#endif

// update lastmem info and write to file (load)
VOID RecordMemRead(VOID * ip, VOID * addr, UINT32 thread_id, VOID *sp, VOID* ebp, VOID* br_taddr, VOID* retAddr, CONTEXT* ctxt)
{
        long sign = 0;
        char valid = 'N';
        //ADDRINT val = 0;
 
        GetLock(&lock, thread_id+1);
        
        if(icount.CountWithREP() == icount_lastmem) 
        {sincelast = 0;}
	else 
        {
		sincelast = icount.CountWithREP() - icount_lastmem - 1;
		icount_lastmem = icount.CountWithREP();
	}


        if(thread_id < 0)
        {
          cout<<"Invalid Thread ID."<<endl;
          PIN_Detach();
          ReleaseLock(&lock);
          return;
        }

#ifdef MULTI_FILE
        if((fptr[thread_id]) == 0)
        {
          sprintf(fileName, "%d", thread_id);
          
          if((fptr[thread_id] = fopen(fileName, "w"))== NULL)
          {
           cout<<"File Creation Error."<<endl;
           PIN_Detach();
           ReleaseLock(&lock);
           return;
          }           
        }
        
        fprintf(fptr[thread_id],"%3u l %p %u\n", thread_id, addr, sincelast);
#endif
         
        //fprintf(stdout,"ip:%p ra:%p sp:%p tid:%3u l addr:%p %u\n", ip, retAddr, sp, thread_id, addr, sincelast);
        //fprintf(stdout,"ip:%16p ra:%16p bt:%12p ebp:%20p sp:%12p tid:%3u l addr:%12p &local:%12p %u\n",
        //                   ip, retAddr, br_taddr, ebp, sp, thread_id, addr, &local, sincelast);
        {
          long* sf_ptr = (long*)ebp; 
          //fprintf(stdout,"gen sf_ptr: %16p\n", sf_ptr);
          
#if 0
          if((unsigned long)sf_ptr >= (unsigned long)sp && 
             (((unsigned long)sf_ptr - (unsigned long)sp) <= SP_BP_DIFF) && \
            ((unsigned long)ip <= (unsigned long)((unsigned long)mainFunAddr + BIN_MARGIN)) && \
             ((unsigned long)ip >= (unsigned long)(((unsigned long)mainFunAddr - BIN_MARGIN)))\
             ) //for valid frame pointer values and IP values w.r.t main()
#endif
          if(((unsigned long)sf_ptr >= (unsigned long)sp) &&
            (((unsigned long)sf_ptr - (unsigned long)sp) <= SP_BP_DIFF) && \
            (RANGE(mainFunAddr)== RANGE(ip))\
            )
            {
              //fprintf(stdout,"*sf_ptr: %16p  *(sf_ptr+1): %16p\n",(void*)sf_ptr[0],(void*)sf_ptr[1]);
              //do_stack_unwind ((void**)sf_ptr);
              getSignature((void**)ip, (void**)sf_ptr,  &sign);
              //use the type to avoid same signature for load/store instructions with same PC value
              //sign = sign ^ INS_TYPE_LOAD;
              if(sign == prev_sign[thread_id])
              {
                sign = sign ^ 'l';
              }  
              prev_sign[thread_id] = sign;
              
              valid = 'V';             
             }
          else //for invalid stack frame pointer values, signature is considered 0
             {
               sign = 0;
               valid = 'I';
               //fprintf(stdout,"*sf_ptr: %16p *(sf_ptr+1): %16p\n", (void*)0x8888, (void*)0x7777);
             }
 
          //fprintf(stdout, "\n#Sign:%20p\n", (void*)sign);
          if(valid == 'V')
          {
           #ifdef STDOUT
           fprintf(stdout, LOAD_OUTPUT_FORMAT);
           #endif
           #ifdef PIPE
           if(logfptr && fpipe)
           {
		long val = 0;
                long type = INS_TYPE_LOAD;
		UINT32 tag = TRACE_TAG;
                UINT32 sign_len = 0;
		UINT32 node_id = 0;                
		LIST_ITERATOR<long> *iter = signList.createIterator(); 
		//fprintf(logfptr, "\n*%d*", (int)signList.getSize()); 
		//fprintf(logfptr,"\n<");
		fwrite(&tag, sizeof(UINT32), 1, fpipe); 
                sign_len = (UINT32)signList.getSize();
		fwrite(&sign_len, sizeof(UINT32), 1, fpipe); 
		while(iter->next())
		{
			//fprintf(logfptr,"%p, ", (void*)iter->currentItem()); 
			val = iter->currentItem();
                        //write ret values list (signature)
		        fwrite(&val, sizeof(long), 1, fpipe); 
		}
		signList.removeAll();
		//fprintf(logfptr,">\n");
		fprintf(logfptr, LOAD_OUTPUT_FORMAT);
                //write refID
		fwrite(&sign, sizeof(long), 1, fpipe); 
                //type
		fwrite(&type, sizeof(long), 1, fpipe);
		//thread id
		fwrite(&thread_id, sizeof(UINT32), 1, fpipe); 
                //node id
		fwrite(&node_id, sizeof(UINT32), 1, fpipe); 
                //address
		fwrite(&addr, sizeof(long), 1, fpipe); 
	   }
           #endif
          }
        }

	++memrefs;

	if(memrefs > MAX_REFS) 
        {
		cout << "Maximum number of memops reached. Stopping instrumentation... " << endl;
    	        //fclose(trace);
    	        #ifdef MULTI_FILE
                 if(fptr)
                 {
                  for(int i=0; i<MAX_THREADS; i++)
                  {
                   if(fptr[i])
                    {fclose(fptr[i]);}
                  }
                 }
                #endif
                #ifdef PIPE
                 //if(logfptr)
                  //{fclose(logfptr); logfptr = NULL;}
                #endif
		PIN_Detach();
	}
        ReleaseLock(&lock); 
}

// Print a memory write record
VOID RecordMemWrite(VOID * ip, VOID * addr, UINT32 thread_id, VOID* sp, VOID* ebp, VOID* br_taddr, VOID* retAddr, CONTEXT* ctxt)
{
        long sign = 0;
        char valid = 'N';
        //ADDRINT val = 0;

        GetLock(&lock, thread_id+1);
	if(icount.CountWithREP() == icount_lastmem) 
        {sincelast = 0;}
	else 
        {
		sincelast = icount.CountWithREP() - icount_lastmem - 1;
		icount_lastmem = icount.CountWithREP();
	}
        
        if(thread_id < 0)
        {
          cout<<"Invalid Thread ID."<<endl;
          PIN_Detach();
          ReleaseLock(&lock);
          return;
        }

     #ifdef MULTI_FILE
        if(fptr[thread_id] == 0)
        {
          sprintf(fileName, "%d", thread_id);

          if((fptr[thread_id] = fopen(fileName, "w"))== NULL)
          {
           cout<<"File Creation Error."<<endl;
           PIN_Detach();
           ReleaseLock(&lock);
           return;
          }
        }
        
        fprintf(fptr[thread_id],"%3u s %p %u\n",thread_id, addr, sincelast);
    #endif
        
        //fprintf(stdout,"ip:%p ra:%p sp:%p tid:%3u s addr:%p %u\n", ip, retAddr, sp, thread_id, addr, sincelast);
        //fprintf(stdout,"ip:%16p ra:%16p bt:%12p ebp:%20p sp:%12p tid:%3u s addr:%12p &local:%12p %u\n", 
        //ip, retAddr, br_taddr, ebp, sp, thread_id, addr, &local, sincelast);
        {
          long* sf_ptr = (long*)ebp;
          //fprintf(stdout,"gen sf_ptr: %16p\n", sf_ptr);
#if 0
             if( (unsigned long)sf_ptr >= (unsigned long)sp && \
                (((unsigned long)sf_ptr - (unsigned long)sp) < SP_BP_DIFF) && \
                ((unsigned long)ip <= (unsigned long)((unsigned long)mainFunAddr + BIN_MARGIN)) && \
                ((unsigned long)ip >= (unsigned long)((unsigned long)mainFunAddr - BIN_MARGIN))\
             ) //for valid frame pointer values and IP values w.r.t main()
#endif
            if((unsigned long)sf_ptr >= (unsigned long)sp &&
            (((unsigned long)sf_ptr - (unsigned long)sp) <= SP_BP_DIFF) && \
            (RANGE(mainFunAddr) == RANGE(ip))\
            )

             {
              //fprintf(stdout,"*sf_ptr: %16p  *(sf_ptr+1): %16p\n",(void*)sf_ptr[0],(void*)sf_ptr[1]);
              //do_stack_unwind ((void**)sf_ptr);
              getSignature((void**)ip, (void**)sf_ptr,  &sign);
              //use the type to avoid same signature for load/store instructions with same PC value
              //sign = sign ^ INS_TYPE_STORE;
              if(sign == prev_sign[thread_id])
              {
                sign = sign ^ 's';
              }  
              prev_sign[thread_id] = sign;
              valid = 'V';             
             }
          else//for invalid stack frame pointer values, signature is considered 0
             {
              sign = 0;
              valid = 'I';
              //fprintf(stdout,"*sf_ptr: %16p *(sf_ptr+1): %16p\n", (void*)0x8888, (void*)0x7777);
             }

          //fprintf(stdout, "\n#Sign:%20p\n", (void*)sign);
          if(valid == 'V')
          {
           #ifdef STDOUT
           fprintf(stdout, STORE_OUTPUT_FORMAT);
           #endif
           #ifdef PIPE 
           if(logfptr && fpipe)
           {
		long val = 0;
                long type = INS_TYPE_STORE;
                UINT32 sign_len = 0;
		UINT32 node_id = 0;
		UINT32 tag = TRACE_TAG;
		LIST_ITERATOR<long> *iter = signList.createIterator();
		//fprintf(logfptr, "\n*%d*", (int)signList.getSize()); 
		fwrite(&tag, sizeof(UINT32), 1, fpipe); 
                sign_len = (UINT32)signList.getSize();
		fwrite(&sign_len, sizeof(UINT32), 1, fpipe); 
		//fprintf(logfptr, "\n<");
		while(iter->next())
		{
			//fprintf(logfptr,"%p, ", (void*)iter->currentItem()); 
			val = iter->currentItem();
                        //write ret values list (signature)
		        fwrite(&val, sizeof(long), 1, fpipe); 
		}
		signList.removeAll();
		//fprintf(logfptr, ">\n");
		fprintf(logfptr, STORE_OUTPUT_FORMAT);
                //write refID
		fwrite(&sign, sizeof(long), 1, fpipe); 
                //type
		fwrite(&type, sizeof(long), 1, fpipe);
		//thread id
		fwrite(&thread_id, sizeof(UINT32), 1, fpipe); 
                //node id
		fwrite(&node_id, sizeof(UINT32), 1, fpipe); 
                //address
		fwrite(&addr, sizeof(long), 1, fpipe); 
	   }
           #endif    
          }
        }

	++memrefs;

	if(memrefs > MAX_REFS) 
        {
		cout << "Maximum number of memops reached. Stopping instrumentation... " << endl;
    	        //fclose(trace);
    	        #ifdef MULTI_FILE
                 if(fptr)
                 {
                  for(int i=0; i<MAX_THREADS; i++)
                  {
                   if(fptr[i])
                    {fclose(fptr[i]);}
                  }
                 }
                #endif
                
                #ifdef PIPE
                 //if(logfptr)
                  //{fclose(logfptr); logfptr = NULL;}
                #endif
		PIN_Detach();
	}
        ReleaseLock(&lock); 
}


VOID ImageLoad(IMG img, VOID * v)
{
   IMG_TYPE imgType = IMG_Type(img);

    //cout<<"\n"<<"!!!!IMAGE LOAD 1!!!!"<<IMG_Name(img)<<"\n";

   if(imgType ==  IMG_TYPE_STATIC || imgType == IMG_TYPE_SHARED)
    {
        RTN mainRtn = RTN_FindByName(img, "main");
        //cout<<"\n"<<"!!!!IMAGE LOAD 2!!!!\n";
  #if 0
        //scan symbol table for main() address 
        for(SYM sym = IMG_RegsymHead(img); SYM_Valid(sym); sym = SYM_Next(sym))
        {    
         //cout<<"\n&&&"<<"Name: "<<SYM_Name(sym);
         
         if( strcmp(SYM_Name(sym).c_str(), "main") == 0) 
          {
            mainFunAddr = (AFUNPTR)SYM_Value(sym);  
            cout<<"\n"<<"!!!!MAIN ADDR FOUND!!!!\n"; 
          }
  
         fprintf(stdout, " Addr: %20p&&&\n",(void*)SYM_Value(sym));
        }
  #endif      
        if (RTN_Valid(mainRtn))//save main()'s address to compare the PC values later
        {
          mainFunAddr = RTN_Funptr(mainRtn);
          //cout<<"\n"<<"!!!!MAIN ADDR FOUND *!!!!\n";
          fprintf(stdout, "\nmain()-addr: %20p", mainFunAddr);
        }
    }
}

// Pin calls this function every time a new rtn is executed
VOID Routine(RTN rtn, VOID *v)
{   
#if 0
    // Allocate a counter for this routine
    RTN_COUNT * rc = new RTN_COUNT;

    // The RTN goes away when the image is unloaded, so save it now
    // because we need it in the fini
    rc->_name = RTN_Name(rtn);
    rc->_image = StripPath(IMG_Name(SEC_Img(RTN_Sec(rtn))).c_str());
    rc->_address = RTN_Address(rtn);
    rc->_icount = 0;
    rc->_rtnCount = 0;
#endif
   ADDRINT fun_addr = RTN_Address(rtn);
   string fun_name = RTN_Name(rtn);
   cout<<"\nfunction name:"<<fun_name;
   fprintf(stdout, "fun addr: %20p", (void*)fun_addr);
   
   if(fun_addr == (ADDRINT)main)
     cout<<"\n"<<"!!!!MAIN CALLED!!!!\n";
    
}

VOID ThreadBegin(UINT32 threadid, VOID * sp, int flags, VOID *v)
{
    GetLock(&lock, threadid+1);
    fprintf(stdout, "thread begin %d sp %p flags %x\n",threadid, sp, flags);
    ReleaseLock(&lock);
}

VOID ThreadEnd(UINT32 threadid, INT32 code, VOID *v)
{
    GetLock(&lock, threadid+1);
    fprintf(stdout, "thread end %d code %d\n",threadid, code);
    ReleaseLock(&lock);
}

VOID ContextChangeCb(THREADID threadid, CONTEXT_CHANGE_REASON reason, const CONTEXT *from, CONTEXT *to, INT32 info, VOID *v)
{
    GetLock(&lock, threadid+1);
    fprintf(stdout, "thread context switch id:%d reason:%d ctx_from:%p ctx_to:%p\n",threadid, reason, from, to);
    ReleaseLock(&lock);
}

// Is called for every instruction and instruments reads and writes
VOID Instruction(INS ins, VOID *v)
{
	if(doinst) {
                
		// instruments loads using a predicated call, i.e.
		// the call happens iff the load will be actually executed
		// (this does not matter for ia32 but arm and ipf have predicated instructions)
		if (INS_IsMemoryRead(ins))
		{
			INS_InsertPredicatedCall(
				ins, IPOINT_BEFORE, (AFUNPTR)RecordMemRead,
				IARG_INST_PTR,
				IARG_MEMORYREAD_EA,
                                IARG_THREAD_ID,
                                IARG_REG_VALUE, REG_STACK_PTR,
                                IARG_REG_VALUE, REG_RBP,
                                IARG_BRANCH_TARGET_ADDR,
                                IARG_RETURN_IP,
                                IARG_CONTEXT,
				IARG_END);
		}

		// instruments stores using a predicated call, i.e.
		// the call happens iff the store will be actually executed
		if (INS_IsMemoryWrite(ins))
		{
			INS_InsertPredicatedCall(
				ins, IPOINT_BEFORE, (AFUNPTR)RecordMemWrite,
				IARG_INST_PTR,
				IARG_MEMORYWRITE_EA,
                                IARG_THREAD_ID, 
                                IARG_REG_VALUE, REG_STACK_PTR,
                                IARG_REG_VALUE, REG_RBP,
                                IARG_BRANCH_TARGET_ADDR,  
                                IARG_RETURN_IP,
                                IARG_CONTEXT,
				IARG_END);
		}
	}
}

VOID Handler(INSTLIB::CONTROL_EVENT ev, VOID * v, CONTEXT * ctxt, VOID * ip, THREADID tid)
{
    switch(ev)
    {
      case INSTLIB::CONTROL_START:
        cout << "Starting instrumentation..." << tid <<endl;
		doinst = true;
		icount_lastmem = icount.CountWithREP();
        break;

      case INSTLIB::CONTROL_STOP:
        cout << "Stopping instrumentation..." << tid <<endl;
		doinst = false;
        break;

      default:
        ASSERTX(false);
        break;
    }
}

VOID Fini(INT32 code, VOID *v)
{
    //fclose(trace);
   printf("......Fini()..........\n");
#ifdef MULTI_FILE
    //close all the opened files
    if(fptr)
    {
      for(int i=0; i<MAX_THREADS; i++)
      { 
        if(fptr[i])
         {fclose(fptr[i]);}
      }
    }
#endif

    if(logfptr)
     {
      fclose(logfptr);
      logfptr = 0;
     }
    
#ifdef PIPE
     if(fpipe)
     {
      fclose(fpipe);
      fpipe = 0;
     }
#endif
    
}


int main(int argc, char *argv[])
{
    PIN_Init(argc, argv);

    //trace = fopen("mem.trace", "w");

#ifdef MULTI_FILE
    for(int i=0; i<MAX_THREADS; i++)
      {fptr[i] = NULL;}
#endif

#ifdef PIPE
     if((fpipe = fopen(PRSD_PIPE, "w"))== NULL)
     {
      printf("pipe %s not present....exiting\n", PRSD_PIPE);
      exit(1);
     }     
#endif

    if((logfptr = fopen("log.txt", "w")) ==  NULL)
    {
      printf("logfile not present....exiting\n");
      exit(1);
    }	
    
    for(int i=0; i<MAX_THREADS; i++)
      {prev_sign[i] = 0;}

    // Activate instruction counter
    icount.Activate();
    icount_lastmem = 1;
    memrefs = 1;

    // Activate alarm, must be done before PIN_StartProgram
    control.CheckKnobs(Handler, 0);

    doinst = false;

    //PIN_AddThreadBeginFunction(ThreadBegin, 0);
    //PIN_AddThreadEndFunction(ThreadEnd, 0);
    //PIN_AddContextChangeFunction(ContextChangeCb, 0);
    PIN_InitSymbols();
 

    INS_AddInstrumentFunction(Instruction, 0);
    // Register Routine to be called to instrument rtn
    //RTN_AddInstrumentFunction(Routine, 0);
    IMG_AddInstrumentFunction(ImageLoad, 0);
    PIN_AddFiniFunction(Fini, 0);

    // Never returns
    PIN_StartProgram();
    
    return 0;
}
