/*
 *   EPRSD template library
 *   Single Author info:
 *   sbudanu SANDEEP BUDANUR RAMANNA
 *
 */

#include<stdio.h>
#include<iostream>
//#include "myqueue.cpp"
using namespace std;

#define MIN(a,b) ((a)<=(b))?(a):(b)
#define abs(a) ((a)>=0)?(a):-(a)

#define STDOUT stdout

typedef short INT16;
typedef int INT32;
typedef long INT64;
typedef unsigned short UINT16;
typedef unsigned int UINT32;
typedef unsigned long UINT64;

typedef enum
{
	CODE_RSD=0,
	CODE_EPRSD
}CODE;

char* code_str[] = {"RSD", "EPRSD"};

/* COMMON class - contains mandatory info of EPRSDs*/
class COMMON
{
	public:
	UINT64* signature;
	UINT64 sign_len;
	UINT64 max_sign_len;
	UINT64 ref_id;
	UINT64 type;
	UINT64 thread_id;
	UINT64 node_id;
	//INT64 value;
	char name[10];

	/* constrcutor */
	COMMON(int max_sign_len)
	{
		//cout<<endl<<"\nCOMMON ctor called";
		this->signature = new UINT64(max_sign_len);
		this->max_sign_len = max_sign_len;
		this->sign_len = 0;
		this->ref_id = 0;
		this->type = 0;
		this->thread_id = 0;
		this->node_id = 0;
		this->name[0] = '\0';
	}

	COMMON(const COMMON& src)
	{
		int i = 0;
		//cout<<endl<<"\nCOMMON copy ctor called";
		this->signature = new UINT64(src.max_sign_len);
		this->max_sign_len = src.max_sign_len;
		this->sign_len = src.sign_len;

		for(i=0; i<this->sign_len; i++)
			this->signature[i] = src.signature[i];

		this->ref_id = src.ref_id;
		this->type = src.type;
		this->thread_id = src.thread_id;
		this->node_id = src.node_id;
		strcpy(this->name, src.name);
	}

	~COMMON()
	{
		//cout<<"\nCOMMON dtor()";
		if(signature)
			delete signature;
	}

	void setData(UINT64* signature, UINT64 sign_len, UINT64 ref_id, UINT64 type, UINT64 thread_id, UINT64 node_id, const char* name)
	{
		int name_len = 0, i = 0, j = 0;

		if(!this->signature){cerr<<"\nsign ptr NULL";}
		for(i=0;i<sign_len; i++)
			this->signature[i] = signature[i];
		this->sign_len = sign_len;
		this->ref_id = ref_id;
		this->type = type;
		this->thread_id = thread_id;
		this->node_id = node_id;
		name_len = min(strlen(name),sizeof(this->name)-1);
		for(j=0; j<name_len; j++)
			this->name[j] = name[j];
		this->name[j] = '\0';
		return;
	}

	void print(FILE* outfptr)
	{
		int i = 0;
		fprintf(outfptr, "\nSign_len = %d", sign_len);
		fprintf(outfptr, "\nSign:");
		for(i=0;i<sign_len;i++)
			fprintf(outfptr, "%p ", signature[i]);

		fprintf(outfptr, "\nref_id = %p",ref_id);
		fprintf(outfptr, "\ntype = %p", type);
		fprintf(outfptr, "\nthread_id = %d", thread_id);
		fprintf(outfptr, "\nnode_id = %d", node_id);
		fprintf(outfptr, "\nname = ");
		if(name)
			fprintf(outfptr, "%s", name);
		return;
	}
};

struct details
{
	INT64 type; //e.g: TYPE_ADDR for address, TYPE_THREAD_ID, TYPE_NODE_ID
	/* start value, length and stride of intra-task values e.g: memory address, destination address, etc.*/
	INT64 start_value;
	INT64 length;
	INT64 stride;
	/* start value, length and stride of inter-task values (thread id) */
	INT64 tid_start_value;
	INT64 tid_length;
	INT64 tid_stride;
	/* start value, length and stride of inter-node values (node id) */
	INT64 node_id_start_value;
	INT64 node_id_length;
	INT64 node_id_stride;

	INT64 loopsize;//no. of LD/ST in a loop iteration
	INT64 loopcount;//no. of loop iterations
	INT64 eprsd_merge_count;
};

template <class T>
class EPRSD_DATA;

/* EPRSD class */
template <class T>
class EPRSD
{
	private:
		CODE code; //if beginning of a loop, it is CODE_EPRSD otherwise CODE_RSD
		struct details det[3];// if more than one stride exists
	public:
		EPRSD<T>* prev;
		EPRSD<T>* next;
		EPRSD<T>* endofloop;
		//EPRSD<T>* left;
		//EPRSD<T>* right;
		//EPRSD<T>* parent;
	public:
		COMMON* cmninfo;//filled by user
		T* userinfo;//defined by user
		template <class U> friend class EPRSD_DATA;

		#if 0
		EPRSD()
		{
			this->cmninfo = NULL;
                        this->userinfo = NULL;
                        this->det[0].length = 1;
                        this->det[0].stride = 0;
                        this->det[0].start_value = 0;
			this->det[0].loopsize = 0;
			this->det[0].loopcount = 0;
			this->det[0].eprsd_merge_count = 0;
                        this->prev = NULL;
                        this->next = NULL;
			this->endofloop = NULL;
                        this->code = CODE_RSD;
		}
		#endif

		EPRSD(COMMON *cmninfo, T* userinfo)
		{
			this->cmninfo = new COMMON(*cmninfo);
			this->userinfo = new T(*userinfo);
                        this->det[0].length = 1;
                        this->det[0].stride = 0;
                        //this->det[0].start_value = 0;
			this->det[0].loopsize = 0;
			this->det[0].loopcount = 0;
			this->det[0].eprsd_merge_count = 0;
                        this->prev = NULL;
                        this->next = NULL;
			this->endofloop = NULL;
                        this->code = CODE_RSD;
			//this->cmninfo = cmninfo;
			//this->userinfo = userinfo;
			this->det[0].start_value = userinfo->getValue();

			this->det[0].tid_start_value = (this->cmninfo)->thread_id;
			this->det[0].tid_length = 1;
			this->det[0].tid_stride = 0;
			this->det[0].node_id_start_value = (this->cmninfo)->node_id;
			this->det[0].node_id_length = 1;
			this->det[0].node_id_stride = 0;
		}

		EPRSD(EPRSD_DATA<T>* data)
		{
			int i=0;
			this->cmninfo = new COMMON(*(data->cmninfo));
			this->userinfo = new T(*(data->userinfo));

			for(i=0;i<data->det_len;i++)
			{
				this->det[i] = data->det[i];
			}
                        this->prev = NULL;
                        this->next = NULL;
			this->endofloop = NULL;
                        this->code = data->code;
		}

		~EPRSD()
		{
			if(cmninfo)
				delete cmninfo;
			if(userinfo)
				delete userinfo;
		}

		void print(FILE* outfptr, INT64* endloop_ref_id, INT32* level_counter, INT32* num_level)
		{
			int i=0;
			int tmp = 0;

			//printf("\n ptr = %p", this);
			if(code == CODE_EPRSD)
			{
				if(endofloop)//for last RSD in EPRSD, endofloop will be NULL
				{
					if(det[0].eprsd_merge_count > 0)
						fprintf(outfptr, "\n\n\nOUTERMOST LOOP COUNT: %d", det[0].eprsd_merge_count+1);
					fprintf(outfptr, "\n\n<EPRSD BEGINS, END: %p LOOPSIZE: %d LOOPCOUNT: %d>",(endofloop->cmninfo)->ref_id, \
													det[0].loopsize, det[0].length);
					level_counter[*num_level] = det[0].loopsize;
					endloop_ref_id[*num_level] = (endofloop->cmninfo)->ref_id;
					(*num_level)++;
				}
			}
			fprintf(outfptr, "\n\ncode: %d", code);
			fprintf(outfptr, "\nstart_value: %p", det[0].start_value);
			fprintf(outfptr, "\nlength: %d", det[0].length);
			fprintf(outfptr, "\nstride: %p", det[0].stride);
			fprintf(outfptr, "\ntid_start_value: %p", det[0].tid_start_value);
			fprintf(outfptr, "\ntid_length: %d", det[0].tid_length);
			fprintf(outfptr, "\ntid_stride: %p", det[0].tid_stride);
			fprintf(outfptr, "\nnode_id_start_value: %p", det[0].node_id_start_value);
			fprintf(outfptr, "\nnode_id_length: %d", det[0].node_id_length);
			fprintf(outfptr, "\nnode_id_stride: %p", det[0].node_id_stride);
			//printf("\neprsd merge: %p", det[0].eprsd_merge_count+1);//n-1 merges produce n eprsd count

			  if(cmninfo)
			    cmninfo->print(outfptr);

			  /*
			  if((*num_level) > 0)
			  {
				printf("*%d*", *num_level);
				if(endloop_ref_id[(*num_level)-1] == cmninfo->ref_id)
				{
					printf("\n<EPRSD ENDS: %p>", endloop_ref_id[(*num_level)-1]);
					(*num_level)--;
				}
			  }
			  */

			  for(i=0;i<(*num_level);i++)
			  {
				//printf("*%d*", *num_level);
				level_counter[i]--;
				if(level_counter[i] == 0)
				{
					fprintf(outfptr, "\n</EPRSD ENDS: %p  LOOPCOUNT: %d>",endloop_ref_id[i], det[0].length);
					//(*num_level)--;
				}
			  }
			  tmp = (*num_level);
			  for(i=0;i<tmp;i++)
			  {
				 if(level_counter[i] == 0)
					(*num_level)--;
			  }
		}

		inline void setLength(INT64 len)
		{ this->det[0].length = len; }
		inline void setStride(INT64 stride)
		{ this->det[0].stride = stride; }
		inline void setStartValue(INT64 start_value)
		{ this->det[0].start_value = start_value; }

		inline INT64 getLength()
		{ return this->det[0].length; }
		inline INT64 getStride()
		{ return this->det[0].stride; }
		inline INT64 getStartValue()
		{ return this->det[0].start_value; }

		inline void setCode(CODE code)
		{ this->code = code; }
		inline CODE getCode()
		{ return this->code; }

		inline INT64 getLoopSize()
		{ return this->det[0].loopsize;	}
		inline void setLoopSize(INT64 size)
		{ this->det[0].loopsize = size; }

		inline INT64 getLoopCount()
		{ return this->det[0].loopcount;}
		inline void setLoopCount(INT64 count)
		{ this->det[0].loopcount = count; }

		inline void setEprsdMergeCount(INT64 count)
		{ this->det[0].eprsd_merge_count = count; }
		inline INT64 getEprsdMergeCount()
		{ return this->det[0].eprsd_merge_count; }

		inline INT64 getThreadId()
		{ return this->det[0].tid_start_value; }
		inline INT64 getThreadIdLength()
		{ return this->det[0].tid_length; }
		inline INT64 getThreadIdStride()
		{ return this->det[0].tid_stride; }
		inline void setThreadId(int tid)
		{ this->det[0].tid_start_value = tid; }
		inline void setThreadIdLength(int len)
		{ this->det[0].tid_length = len; }
		inline void setThreadIdStride(int stride)
		{ this->det[0].tid_stride = stride; }

		inline INT64 getNodeId()
		{ return this->det[0].node_id_start_value; }
		inline INT64 getNodeIdLength()
		{ return this->det[0].node_id_length; }
		inline INT64 getNodeIdStride()
		{ return this->det[0].node_id_stride; }
};

/*EPRSD compare function */
typedef bool(*cmp_fun)(COMMON*, void*, COMMON*, void*);
/*
class EPRSD_ITERATOR
{
	EPRSD<T>* cur;
	public:
	EPRSD_ITERATOR()
	{
		cur = lis;
	}

	~EPRSD_ITERATOR()
	{

	}
}
*/

//template <class T, template <class M> class MATCH, template <class G> class MERGE>
//class EPRSD_COMPRESSOR;

//template <class T, template <class M> class MATCH, template <class G> class MERGE>
//class ITERATOR;
template <class T>
class EPRSD_DATA
{
		public:
		CODE code; //if beginning of a loop, it is CODE_EPRSD otherwise CODE_RSD
		INT16 det_len; //no. of details structure used
		struct details det[3];// if more than one stride exists
		COMMON* cmninfo;//filled by user
		T* userinfo;//defined by user

		public:
		EPRSD_DATA()
		{
			code = CODE_RSD;
			det_len = 1;
			cmninfo = NULL;
			userinfo = NULL;
		}

		//creates EPRSD_DATA from EPRSD class
		void copy(EPRSD<T> *src)
		{
			code = src->code;
			det_len = 1; //TODO: modify this
			det[0] = src->det[0];
			det[1] = src->det[1];
			det[2] = src->det[2];
			//only pointer is copied
			cmninfo = src->cmninfo;
			userinfo = src->userinfo;
		}

		//creates EPRSD_DATA from details, cmn and userinfo
		void copy(CODE code, struct details* det, int det_len, COMMON* cmninfo, T* userinfo)
		{
			int i = 0;
			this->code = code;
			for(i=0;i<det_len;i++)
			{
				this->det[i] = det[i];
			}
			//data is copied, so allocate memory
			this->cmninfo = new COMMON(*cmninfo);
			this->userinfo = new T(*userinfo);
		}
};

/* EPRSD Compressor class*/
template <class T, template <class M> class MATCH, template <class G> class MERGE>
class EPRSD_COMPRESSOR
{
	private:
		EPRSD<T> *listhead;
		EPRSD<T> *listtail;
		EPRSD<T> *l_match_begin;//starting node of left matching portion
		EPRSD<T> *l_match_end;//ending node of left matching portion
		EPRSD<T> *r_match_begin;//starting node of right matching portion
		EPRSD<T> *r_match_end;//ending node of right matching portion
		EPRSD<T> *latest_eprsd;//ending node of right matching portion
		/* for inter task compression */
		EPRSD<T> *it_l_match_begin;
		EPRSD<T> *it_l_match_end;
		EPRSD<T> *it_r_match_begin;
		EPRSD<T> *it_r_match_end;

	public:
		INT32 level_counter[10];
		INT64 endloop_ref_id[10];
		INT32 num_level;

	private:
		template <class _T, template <class _M> class _MATCH, template <class _G> class _MERGE>
		friend class ITERATOR;
		/*
		void createEPRSDIterator(EPRSD_ITERATOR *iter)
		{

		}
		void getNextEPRSD(EPRSD_ITERATOR *iter, EPRSD_DATA<T> *data)
		{

		}
		*/
		/* returns the pointer to matched EPRSD otherwise NULL */
		bool isSignMatch(COMMON* src, COMMON* dest)
		{
			int i=0;
			//cout<<"\nCalling cmp function";
			//if reference id donot match , return false
			#if 0
			if(src->ref_id == dest->ref_id)
			{
				return true;
			}
			else
			{
				return false;
			}
			#endif
			#if 1
			if(src->ref_id != dest->ref_id)
				return false;
			else //compare signature
			{
				if(src->sign_len != dest->sign_len)
					return false;
				else
				{
					for(i=0;i<src->sign_len;i++)
					{
						if(src->signature[i] != dest->signature[i])
							return false;
					}
				}
			}
			return true;
			#endif

		}
		/* loose match needed for add()*/
                EPRSD<T>* findLastMatch(EPRSD<T>* neweprsd)
                {
                        EPRSD<T>* cur = NULL;
                        EPRSD<T>* next = NULL;
                        //list empty
                        if(!listhead || !listtail || !neweprsd)
                                return NULL;

                        cur = listtail;
                        while(cur)
                        {
				//if(cur->getCode() == CODE_RSD && neweprsd->getCode() == CODE_RSD)//compare only RSDs, RSD and EPRSD don't match
				{
                                 //if( match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo) )
				 if( 	isSignMatch(cur->cmninfo, neweprsd->cmninfo)\
					&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) )
                                 {
                                        return cur;
                                 }
				}
                                cur = cur->prev;
                        }
                        return NULL;
                }

		EPRSD<T>* findExactMatch(EPRSD<T>* neweprsd)
		{
                        EPRSD<T>* cur = NULL;
                        EPRSD<T>* next = NULL;
			//fprintf(STDOUT, "\nfindExactMatch called - %p", neweprsd->cmninfo->ref_id);
                        //list empty
                        if(!listhead || !listtail || !neweprsd)
                                return NULL;
                        cur = neweprsd->next;
                        while(cur)
			{
                                 //if( match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo) &&
				 if( 	isSignMatch(cur->cmninfo, neweprsd->cmninfo)\
					&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) &&
				     (cur->getCode() == neweprsd->getCode()) &&
				     (cur->getLoopSize() == neweprsd->getLoopSize())
					)
				//  (((cur->endofloop)->cmninfo)->ref_id == ((neweprsd->endofloop)->cmninfo)->ref_id)
				 {
					//printf("sign, code, loopsize matches");
					if(cur->endofloop && neweprsd->endofloop)
					{
						if ( ((cur->endofloop)->cmninfo)->ref_id == ((neweprsd->endofloop)->cmninfo)->ref_id )
						{
							//printf("$");
							return cur;
						}
					}
					else if(!cur->endofloop && !neweprsd->endofloop)
					{
						//printf("$$");
						return cur;
					}
					else
					{
						fprintf(STDOUT, "\nendofloop-one is NULL other is not NULL");
					}
				 }
				 #if 0
				 else
				 {
                                     printf("<<\n%d %d %d>>", match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo),\
				     (cur->getCode() == neweprsd->getCode()),\
				     (cur->getLoopSize() == neweprsd->getLoopSize()));
				 }
				 #endif
				 cur = cur->next;
			}
			//printf("#");
			return NULL;
		}


		EPRSD<T>* intertask_findExactMatch(EPRSD<T>* neweprsd)
		{
			#if 1
                        EPRSD<T>* cur = NULL;
                        EPRSD<T>* next = NULL;
			//fprintf(STDOUT, "\nfindExactMatch called - %p", neweprsd->cmninfo->ref_id);
                        //list empty
                        if(!listhead || !listtail || !neweprsd)
                                return NULL;
                        cur = listhead;
                        while(cur)
			{
				 if( isSignMatch(cur->cmninfo, neweprsd->cmninfo)&&
				     MATCH<T>::intertask_compare(cur->userinfo, neweprsd->userinfo) &&
				     (cur->getCode() == neweprsd->getCode()) &&
				     (cur->getLoopSize() == neweprsd->getLoopSize()) //&&
				     //(cur->getLoopCount() == neweprsd->getLoopCount()) &&
				     //(cur->getStride() == neweprsd->getStride())
				   )
				 {
					fprintf(STDOUT, "\nintertask match found #1");
					break;
				 }
				 cur = cur->next;
			}
			#endif
			return cur;
		}


		bool isEqual(EPRSD<T>* cur, EPRSD<T>* neweprsd)
		{

			 //if( match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo) &&
				 if( 	isSignMatch(cur->cmninfo, neweprsd->cmninfo)\
					&& MATCH<T>::compare(cur->userinfo, neweprsd->userinfo) &&
                                     (cur->getCode() == neweprsd->getCode()) &&
                                     (cur->getLoopSize() == neweprsd->getLoopSize())
                                        )
                                //  (((cur->endofloop)->cmninfo)->ref_id == ((neweprsd->endofloop)->cmninfo)->ref_id)
                                 {
                                        if(cur->endofloop && neweprsd->endofloop)
                                        {
                                                if ( ((cur->endofloop)->cmninfo)->ref_id == ((neweprsd->endofloop)->cmninfo)->ref_id )
                                                {
                                                        return true;
                                                }
                                        }
                                        else if(!cur->endofloop && !neweprsd->endofloop)
                                        {
                                                return true;
                                        }
                                 }
				 else
				 {
                                     //fprintf(STDOUT,"<<\nisEqual %d %d %d>>", match_fun(cur->cmninfo, cur->userinfo, neweprsd->cmninfo, neweprsd->userinfo),\
				     (cur->getCode() - neweprsd->getCode()),\
				     (cur->getLoopSize() -  neweprsd->getLoopSize()));
				 }
			return false;
		}

		void addTail(EPRSD<T>* neweprsd)
		{
			//neweprsd->setCode(CODE_RSD);
			if(!listhead)
	                {
        	                listhead = neweprsd;
                	        listtail = neweprsd;
                	        neweprsd->prev = NULL;
                	        neweprsd->next = NULL;
	                }
        	        else
                	{
                        	listtail->next = neweprsd;
	                        neweprsd->prev = listtail;
        	                listtail = neweprsd;
                	        neweprsd->next = NULL;
	                }
		}
		void intertask_merge(EPRSD<T>*, EPRSD<T>*);
		void intertask_mergeLeftRightPortions();
		void mergeLeftRightPortions();
		void printWindow(FILE* outfptr, EPRSD<T>* start, EPRSD<T>* end)
		{
			EPRSD<T>* cur;
			//fprintf(STDOUT, "\nprintWindow\n");
			if(!start || !end)
				return;
			cur = start;
			while(cur && cur != end->next)
			{
				//printf("\n*%p*", cur);
				cur->print(outfptr);
				cur = cur->next;
			}
		}

	public:
		//template <class T_, template <class M_> class MATCH_, template <class G_> class MERGE_>
		class ITERATOR
		{
			private:
			EPRSD_COMPRESSOR *cmprsr;
			EPRSD<T> *cur;
			EPRSD_DATA<T> *data;

			public:
			ITERATOR(EPRSD_COMPRESSOR<T, MATCH, MERGE>* ecr)
			{
				cmprsr = ecr;
				cur = NULL;
				data = new EPRSD_DATA<T>();
			}

			~ITERATOR()
			{
				delete data;
			}

			void begin()
			{
				cur = cmprsr->listhead;
			}

			EPRSD_DATA<T>* getNext()
			{
				EPRSD<T> *tmp = NULL;
				tmp = cur;
				if(cur)
					cur = cur->next;
				//return tmp;
				if(tmp)
				{
					data->copy(tmp);//copy EPRSD data portion
					return data;
				}
				return NULL;
			}
		};

		EPRSD_COMPRESSOR()
		{
			int i=0;
			this->listhead = NULL;
			this->listtail = NULL;
			this->l_match_begin = NULL;
			this->l_match_end = NULL;
			this->r_match_begin = NULL;
			this->r_match_end = NULL;
			this->it_l_match_begin = NULL;
			this->it_l_match_end = NULL;
			this->it_r_match_begin = NULL;
			this->it_r_match_end = NULL;
			this->latest_eprsd = NULL;
			this->num_level = 0;
			for(i=0;i<10;i++)
			{
				this->level_counter[i] = 0;
				this->endloop_ref_id[i] = 0;
			}
		}

		void add(COMMON* cmn, T* newentry);
		void add(EPRSD_DATA<T> *data);
		void finalize();

		void print(FILE* outfptr)
		{
			EPRSD<T>* cur = NULL;
			if(!listhead)
				return;
			cur = listhead;
			while(cur)
			{
				cur->print(outfptr, endloop_ref_id, level_counter, &num_level);
				cur = cur->next;
			}
		}
};

template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::intertask_mergeLeftRightPortions()
{
	EPRSD<T> *cur = NULL;
	EPRSD<T> *right_cur = NULL;
	EPRSD<T> *nxt = NULL;
	INT64 stride = 0;
	INT64 length = 0;
	INT64 count = 0;
	//EPRSD is already present
	if(it_l_match_begin->getCode() == CODE_EPRSD && it_l_match_end->getCode() == CODE_EPRSD)//beginning of loop already marked, so just add the length
	{
		//update the length of RSDs
        	cur = it_l_match_begin;
		right_cur = it_r_match_begin;
	        while(cur != it_l_match_end->next)
		{
			intertask_merge(cur, right_cur);
			MERGE<T>::intertask_mergecb(cur->userinfo, right_cur->userinfo);
			cur = cur->next;
			right_cur = right_cur->next;
			//compute the EPRSD length, # of RSDs
			count++;
		}

		if(it_l_match_begin->getLoopSize() != count)
		{
			fprintf(STDOUT, "ERROR: intertask EPRSD merging - count mismatch ");
		}

		//adjust the head/tail before deleting right portion
		//l_match_end->next = r_match_end->next;
	        if(listtail == it_r_match_end)
			listtail = it_r_match_begin->prev;
		//free the right portion
        	cur = it_r_match_begin;
	        while(cur)
        	{
                	nxt = cur->next;
	                delete cur;
        	        cur = nxt;
	        }
		//fprintf(STDOUT, "\nMerge LR Update");
	}
	return;
}


template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::mergeLeftRightPortions()
{
	EPRSD<T> *cur = NULL;
	EPRSD<T> *right_cur = NULL;
	EPRSD<T> *nxt = NULL;
	INT64 stride = 0;
	INT64 length = 0;
	INT64 count = 0;
	//EPRSD is already present
	//if(l_match_begin->parent && l_match_end->parent)//RSDs have non NULL parent if they are under an EPRSD
	if(l_match_begin->getCode() == CODE_EPRSD && l_match_end->getCode() == CODE_EPRSD)//beginning of loop already marked, so just add the length
	{
		//update the length of RSDs
        	cur = l_match_begin;
		right_cur = r_match_begin;
	        while(cur != l_match_end->next)
		{
			length = cur->getLength() + right_cur->getLength();
			//printf("$%d ", length);
			cur->setLength(length);
			if(cur && right_cur)
				stride = (abs( ((cur->userinfo)->getValue()) - ((right_cur->userinfo)->getValue()) ))/((cur->getLength()) - 1);
			cur->setStride(stride);
				MERGE<T>::mergecb(cur->userinfo, right_cur->userinfo);
			cur = cur->next;
			right_cur = right_cur->next;
			//compute the EPRSD length, # of RSDs
			count++;
		}
		//if((l_match_begin->getRoot())->getLength() != count)
		//	fprintf(STDOUT, "EPRSD merging - count mismatch");
		if(l_match_begin->getLoopSize() != count)
		{
			fprintf(STDOUT, "EPRSD merging - count mismatch");
		}

		//adjust the head/tail before deleting right portion
		l_match_end->next = r_match_end->next;
	        if(listtail == r_match_end)
			listtail = l_match_end;
		//free the right portion
        	cur = r_match_begin;
	        while(cur)
        	{
                	nxt = cur->next;
	                delete cur;
        	        cur = nxt;
	        }
		//fprintf(STDOUT, "\nMerge LR Update");
	}
	else
	{
		//fprintf(STDOUT, "\nMerge LR");
		//mark as EPRSD

		l_match_begin->setCode(CODE_EPRSD);
		l_match_end->setCode(CODE_EPRSD);
		l_match_begin->endofloop = l_match_end;
		//update the length of RSDs
        	cur = l_match_begin;
		right_cur = r_match_begin;
		count = 0;
	        while(cur != l_match_end->next)
		{
			//all nodes point to their parent EPRSD node
			length = cur->getLength() + right_cur->getLength();
			//printf("$%d ", length);
			cur->setLength(length);
			if(cur && right_cur)
				stride = (abs( ((cur->userinfo)->getValue()) - ((right_cur->userinfo)->getValue()) ))/((cur->getLength()) - 1);
			cur->setStride(stride);
				MERGE<T>::mergecb(cur->userinfo, right_cur->userinfo);
			cur = cur->next;
			right_cur = right_cur->next;
			cur->setCode(CODE_EPRSD);
			//compute the EPRSD length, # of RSDs
			count++;
		}
		(l_match_begin)->setLoopSize(count);

		//adjust the head/tail before deleting right portion
		l_match_end->next = r_match_end->next;
		if(listtail == r_match_end)
			listtail = l_match_end;

		//free the right portion
		cur = r_match_begin;
		while(cur)
		{
			nxt = cur->next;
			delete cur;
			cur = nxt;
		}

		latest_eprsd = l_match_begin;
		//fprintf(STDOUT, "\nMerge LR end - %p", latest_eprsd);
	}
	return;
}

template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::add(COMMON* cmn, T* newentry)
{
	bool matched = false;
	INT64 stride = 0;
	EPRSD<T> *neweprsd = new EPRSD<T>(cmn, newentry);
	EPRSD<T> *match = NULL;
	//fprintf(STDOUT, "\nadd 1");

	while(1)
	{
		//printf("%d %d %d %d", l_match_begin, l_match_end, r_match_begin, r_match_end);
		if(l_match_begin == NULL)//no match found yet
		{
			//fprintf(STDOUT, "\nl_match_begin == NULL");
			match = findLastMatch(neweprsd);
			if(match)//if match found
			{
			   //fprintf(STDOUT, "\nmatch found");
			   if(l_match_begin == NULL)
			   {
				l_match_begin = match;
			   }
			   l_match_end = match;
			   //add to window and store the postition as right window's starting point
			   //TODO: check the window overflow before adding
			   addTail(neweprsd);
			   if(r_match_begin == NULL)
        	           {
                	        r_match_begin = listtail;
	                   }
        	           r_match_end = listtail;
			}
			else //if match not found, add to window
			{
			   addTail(neweprsd);
			}
			break;
		}
		else//some match was found earlier
		{
			//fprintf(STDOUT, "\nl_match_begin != NULL");
			if(l_match_end->next != r_match_begin)
			{
				//fprintf(STDOUT, "\nl_match_end->next != r_match_begin");
				//if(l_match_end->getCode() == CODE_RSD)
				//{
				 //matching continues
				 //if(match_fun((l_match_end->next)->cmninfo, (l_match_end->next)->userinfo, neweprsd->cmninfo, neweprsd->userinfo))
				 if( 	isSignMatch((l_match_end->next)->cmninfo, neweprsd->cmninfo)\
					&& MATCH<T>::compare((l_match_end->next)->userinfo, neweprsd->userinfo))
				 {
					l_match_end = l_match_end->next;
					addTail(neweprsd);
					r_match_end = listtail;
					//optionally can check for pattern match here again
					break;
				 }
				 else//matching stops
				 {
					 //TODO: increment the length and update the strides ONLY of matched eprsds though there was some mismatch
					 //note the difference - *no EPRSD is created here* grouping the RSDs, only length/stride are updated
					 //reset the begin and end points of l and r windows
				 	 l_match_begin = l_match_end = NULL;
		                         r_match_begin = r_match_end = NULL;
        		                 //reset the window and use the new trace node to match from scratch
                		        // goto FRESH_MATCH;
					continue;
				 }
			}
			else//l_match_end is just before r_match_begin
			{
				//fprintf(STDOUT,"\n Pattern Matched RSD case:");
				//fprintf(STDOUT,"$$$$%p %p$$$$",l_match_begin->cmninfo->ref_id, l_match_end->cmninfo->ref_id);
				//merge the left and right portions, *new EPRSD is created here*
				mergeLeftRightPortions();
				//free the memory of right portion nodes
				l_match_begin = l_match_end = NULL;
		                r_match_begin = r_match_end = NULL;
			}//else ends
		}//else ends
	}//while ends
}

template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::intertask_merge(EPRSD<T>* match, EPRSD<T>* neweprsd)
{
	int length = 1, stride = 0;
	length = match->getThreadIdLength() + neweprsd->getThreadIdLength();
	match->setThreadIdLength(length);
	stride = (abs( (match->getThreadId()) - (neweprsd->getThreadId()) )); // /((match->getThreadIdLength()) - 1);
	match->setThreadIdStride(stride);
	return;
}

template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::add(EPRSD_DATA<T> *data)
{
	EPRSD<T> *match = NULL;
	int growing = 0;
	//if it is an RSD, just merge
	if(!data)
		return;

	EPRSD<T> *neweprsd = new EPRSD<T>(data);

	while(1)
	{
		match = intertask_findExactMatch(neweprsd);
		if(match)
		{
			intertask_merge(match, neweprsd);
			delete neweprsd; /* neweprsd not needed after merging */
			#if 0
			fprintf(STDOUT, "\nInter task - match found #2");
			if(match->getCode() == CODE_RSD)/* immediately merge the RSDs*/
			{
				intertask_merge(match, neweprsd);
				delete neweprsd; /* neweprsd not needed after merging */
			}
			else if(match->getCode() == CODE_EPRSD)/* add to tail and wait for complete match sequence by building window*/
			{
				addTail(neweprsd);
				if(it_l_match_begin == NULL)/* first item of EPRSD sequence */
				{
					if(match->endofloop) /* first node of EPRSD points to the end */
					{
					 it_l_match_begin = match;
					 it_l_match_end = match;
					 it_r_match_begin = neweprsd;
					 it_r_match_end = neweprsd;
					}
					else
					{
						fprintf(STDOUT, "\nERROR: intertask - beginning of EPRSD has NULL endofloop");
					}
				}
				else /* subsequent items of EPRSD sequence */
				{
					/* matching continues */
					if(match == it_l_match_end->next)
					{
						fprintf(STDOUT, "\nintertask window growing...");
						it_l_match_end = match;
						it_r_match_end = neweprsd;
						growing = 1;
					}
					else /* matching stops, reset window pointers */
					{
						fprintf(STDOUT, "\nintertask window stops...");
						it_l_match_begin = NULL;
						it_r_match_begin = NULL;
						it_l_match_end = NULL;
						it_r_match_end = NULL;
						growing = 0;
					}
					/* merge the EPRSD sequence */
					/* EPRSD sequence completed */
					if(growing == 1)
					{
						if(isSignMatch((it_l_match_begin->endofloop)->cmninfo, it_l_match_end->cmninfo ))
						{
							fprintf(STDOUT, "\n intertask EPRSD merging");
							//intertask_mergeLeftRightPortions();
							/* reset the pointers as they are no more required */
							it_l_match_begin = NULL;
							it_r_match_begin = NULL;
							it_l_match_end = NULL;
							it_r_match_end = NULL;
							growing = 0;
						}
					}
				}//else ends

			}//else if ends
			else
			{
				fprintf(STDOUT, "ERROR: Invalid code.....");
			}
			#endif
		}//if(match) ends
		else
		{
			fprintf(STDOUT, "\nInter task - match NOT found");
			addTail(neweprsd);
		}
		break;
	}//while ends

	//if an EPRSD, add as tail and look for the complete sequence match
	//when whole sequence is matched, merge the EPRSD
}

template <class T, template <class M> class MATCH, template <class G> class MERGE>
void EPRSD_COMPRESSOR<T, MATCH, MERGE>::finalize()
{
	EPRSD<T> *match = NULL;
	EPRSD<T> *start = NULL;
	EPRSD<T> *neweprsd = NULL;
	EPRSD<T> *cur = NULL;
	EPRSD<T> *right_cur = NULL;
	EPRSD<T> *nxt = NULL;
	INT64 stride = 0;
	INT64 length = 0;
	INT64 count = 0;
	fprintf(STDOUT, "\n\n*****************finalize called*********************");
	l_match_begin = l_match_end = NULL;
        r_match_begin = r_match_end = NULL;

	start = listhead;
	neweprsd = start;

	while(neweprsd)
	{
		//fprintf(STDOUT, "\nneweprsd: %p", neweprsd->cmninfo->ref_id);
		//handle RSD and EPRSDs differently
		if(neweprsd->getCode() == CODE_RSD)
		{
			match = findExactMatch(neweprsd);
			while(match)
			{
			 	 //fprintf(STDOUT, "\nmatch RSD: %p", match->cmninfo->ref_id);
				 //getchar();
				 //fprintf(STDOUT, "ptrs: %p %p", neweprsd, match);
				 //merge matched entries, assign eprsd being searched and matched into Left & Right portions respectively
				 length = neweprsd->getLength() + match->getLength();
				 neweprsd->setLength(length);
        	                 //stride = (abs( ((neweprsd->userinfo)->getValue()) - ((match->userinfo)->getValue()) ))/((neweprsd->getLength()) - 1);
                	         neweprsd->setStride(stride);
				 //remove the matched node

				 if(listhead != match)
					 (match->prev)->next = match->next;
				 else
					listhead = match->next;

				 if(listtail !=  match)
					 (match->next)->prev = match->prev;
				 else
					listtail = match->prev;

				 delete match;
			 	 match = findExactMatch(neweprsd);
			}
			//printf("\nskipping RSD");
			neweprsd = neweprsd->next;
			continue;
		}
		else if(neweprsd->getCode() == CODE_EPRSD)
		{
			match = findExactMatch(neweprsd);
			//getchar();
			while(match)
			{
			 //fprintf(STDOUT, "\nmatch EPRSD: %p", match->cmninfo->ref_id);
			 //getchar();
			 //fprintf(STDOUT, "ptrs: %p %p", neweprsd, match);
			 //merge matched entries, assign eprsd being searched and matched into Left & Right portions respectively

			 	if(l_match_begin == NULL)
					l_match_begin = neweprsd;
				 l_match_end = neweprsd->endofloop;
				 if(r_match_begin == NULL)
					r_match_begin = match;
				 r_match_end = match->endofloop;

			 //update the length of RSDs
	                cur = l_match_begin;
        	        right_cur = r_match_begin;
			count = 0;
                	 while(cur != l_match_end->next && right_cur != r_match_end->next)
	                 {
        	                length = cur->getLength() + right_cur->getLength();
                        	cur->setLength(length);
				#if 0
	                        if(cur->getCode() == CODE_RSD && right_cur->getCode() == CODE_RSD )//stride changes only for RSDs
				{
        	                        stride = (abs( ((cur->userinfo)->getValue()) - ((right_cur->userinfo)->getValue()) ))/((cur->getLength()) - 1);
                	        	cur->setStride(stride);
				}
				#endif
				cur->setEprsdMergeCount(cur->getEprsdMergeCount()+1);
                        	cur = cur->next;
	                        right_cur = right_cur->next;
				//update how many times an EPRSD was matched and merged
        	                //compute the EPRSD length, # of RSDs
                	        count++;
	                 }
        	        //if((l_match_begin->getRoot())->getLength() != count)
                	//      fprintf(STDOUT, "EPRSD merging - count mismatch");
	                if(l_match_begin->getLoopSize() != count)
			{
        	          fprintf(STDOUT, "EPRSD merging - count mismatch %d %d",l_match_begin->getLoopSize(), count);
			}

			if(r_match_begin->prev)
				(r_match_begin->prev)->next = r_match_end->next;
			if(r_match_end->next)
				(r_match_end->next)->prev = r_match_begin->prev;

        	        if(listtail == r_match_end)
                	        listtail = r_match_begin->prev;
			if(listhead == r_match_begin)
				listhead = r_match_begin->prev;

			r_match_begin->prev = NULL;
			r_match_end->next = NULL;

	                //free the right portion
        	        cur = r_match_begin;
                	 while(cur)
	                 {
	                        nxt = cur->next;
        	                delete cur;
                	        cur = nxt;
	                 }
			 //search again
			 match = findExactMatch(neweprsd);
			 l_match_begin = l_match_end = NULL;
			 r_match_begin = r_match_end = NULL;
			}//while(match) ends
			//fprintf(STDOUT, "\nJumping to the end of EPRSD");
			if(neweprsd->getCode() == CODE_RSD)
				neweprsd = neweprsd->next;
			else if(neweprsd->getCode() == CODE_EPRSD)
				neweprsd = (neweprsd->endofloop)->next;

			continue;
		 }//else if CODE_EPRSD ends
		}
	return;
}

#if 0
template<class T>
void EPRSD_COMPRESSOR<T>::finalize()
{
	EPRSD<T> *match = NULL;
	EPRSD<T> *start = NULL;
	EPRSD<T> *neweprsd = NULL;
	printf("\n\n*****************finalize called*********************");
	l_match_begin = l_match_end = NULL;
        r_match_begin = r_match_end = NULL;

	start = listhead;
	neweprsd = start;

	while(neweprsd)
	{
		printf("\nneweprsd: %p", neweprsd->cmninfo->ref_id);
		match = findExactMatch(neweprsd);
		//getchar();
		if(match)
		{
		 printf("\nmatch: %p", match->cmninfo->ref_id);
		 getchar();
		 #if 0
		 printf("ptrs: %p %p", neweprsd, match);
		 printf("code: %s %s", code_str[neweprsd->getCode()], code_str[match->getCode()]);
		 printf("ptrs: %p %p", neweprsd, match);
		 printf("code: %s %s", code_str[neweprsd->getCode()], code_str[match->getCode()]);
		 #endif
		 printf("ptrs: %p %p", neweprsd, match);

		if(l_match_begin && l_match_end && r_match_begin && r_match_end)
		 {
			//is new match continuous
			if(isEqual(match, r_match_end->next))// (match == r_match_end->next))
			{
				//contents
				(match == r_match_end->next);

			}
			else if(!isEqual(match, r_match_end->next))
			{
				printf("\nmatching stops");
				printf("\nmatching stops");
				l_match_begin = l_match_end = NULL;
			        r_match_begin = r_match_end = NULL;
				start = start->next;
				neweprsd = start;
				continue;
			}
		 }
			 	if(l_match_begin == NULL)
					l_match_begin = neweprsd;
				 l_match_end = neweprsd;
				 if(r_match_begin == NULL)
					r_match_begin = match;
				 if(r_match_end == NULL)
					r_match_end = match;
				 else //contensts are equal, pointer may not be
				 {
				 	r_match_end = r_match_end->next;
				 }
				 #if 1
				 if(l_match_end->next == r_match_begin)
				 {
					printf("\nPerfect Match....Merge\n");
				 }
				 printf("\nL:%p %p", l_match_begin->cmninfo->ref_id, l_match_end->cmninfo->ref_id);
				 printf("\nR:%p %p", r_match_begin->cmninfo->ref_id, r_match_end->cmninfo->ref_id);
				 #endif

		}
		neweprsd = neweprsd->next;
	}
	return;
}
#endif
/* abstract class making getValue() and setValue() pure virtual functions - user must define them*/
class INFO
{
	protected:
	UINT64 value;

	public:
	INFO(UINT64 addr)
	{
		//cout<<"\nINFO ctor";
		value = addr;
	}
	INFO()
	{
		value = 0;
	}
	~INFO()
	{
		//cout<<"\nINFO dtor";
	}
	virtual INT64 getValue() = 0;
	virtual void setValue(void* val) = 0;
	//must overload - operator
};

